; $Id: dm_load_macs.pro,v 1.31 2016/03/07 21:04:01 ymqiu Exp $
;#######################################################################
; 
; NAME:
;  dm_load_macs
;
; PURPOSE:
;  load macs file
;  
; CATEGORY:
;  dcs_mslice
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  March, 2016
; 
; HISTORY:
;  02/2013 V1
;  07/2013 Add magnetic field and specified data; take into account the temperture unit.
;  08/2013 Allow dm_load_macs to read a file content string array
; 
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

;read macs data file
;input:
;  file:            full file name string(s), or a string array of a full file content
;  group_leader:    parent ID
;  answeryes/no:    suppress the prompt question with default answer yes/no
;  checkonly:       if set, will not promp any question about inconsistent number of columns, and will not check for negative detector counts
;  firstlineonly:   if set, will read only the first line of data
;  string:          if set, output data as string array, exactly the same as the content in the data file
;output:
;  data:            data array [nrow,ncol], if string keyword is set, a string array, otherwise a double array
;  info:            description of columns, lower case output, if string is set, keep the original letter case
;  all_neg:         flag to indicate that all detectors contain -1, no valid data whatsoever
;  comment:         comment
;  deteff:          detector efficiency [2,ndet], [0,*] for spec detector, [1,*] for diff detector
;  error:           error flag
;  filename:        filename
;  flipstring:      flip field string, a three-element string array
;  header:          file header string array  
;  instrname:       instrument name, lower case output
;  latticeori:      [ux,uy,uz,vx,vy,vz]
;  latticeparm:     [a,b,c,alpha,beta,gamma]
;  maspacing:       [monochromator,analyser] spacing
;  npoints:         #npoints value in the header of the first file
;  temperatureunit: the #TemperatureUnits field value
;  trailer:         file trailer string array, may be empty
pro dm_load_macs,file,data,info,all_neg=all_neg,answerno=answerno,answeryes=answeryes,comment=comment,checkonly=checkonly,deteff=deteff,error=error,filename=filename,firstlineonly=firstlineonly,flipstring=flipstring,$
    group_leader=group_leader,header=header,instrname=instrname,latticeori=latticeori,latticeparm=latticeparm,maspacing=maspacing,npoints=npoints,string=string,temperatureunit=temperatureunit,trailer=trailer
  	if n_elements(comment) ne 0         then tmp = temporary(comment) 
  	if n_elements(data) ne 0            then tmp = temporary(data)   
  	if n_elements(info) ne 0            then tmp = temporary(info)   
  	if n_elements(header) ne 0          then tmp = temporary(header)
  	if n_elements(trailer) ne 0         then tmp = temporary(trailer)  
  	if n_elements(instrname) ne 0       then tmp = temporary(instrname)
  	if n_elements(all_neg) ne 0         then tmp = temporary(all_neg)
  	if n_elements(temperatureunit) ne 0 then tmp = temporary(temperatureunit)
  	if n_elements(filename) ne 0        then tmp = temporary(filename)
  	if arg_present(npoints)             then npoints     = 0
  	if arg_present(maspacing)           then maspacing   = fltarr(2)
  	if arg_present(latticeparm)         then latticeparm = fltarr(6)
  	if arg_present(latticeori)          then latticeori  = fltarr(6)
  	if arg_present(deteff)              then deteff      = fltarr(2,20)
  	if arg_present(flipstring)          then flipstring  = ['','','']
  	zerostring = ['out','beo','a','nsf']      ;strings for 0
  	onestring  = ['in','be','b','sf']         ;strings for 1
  	twostring  = ['hopg']                     ;strings for 2
  	tmp=' ' & header=' '
  	error = 0b
  	if total(stregex(file,'#ICERepositoryInfo',/fold_case,/boolean)) gt 0 then begin ;file is actually a file content string array
  	   ind_column = where(stregex(file,'#columns',/boolean,/fold_case),count_col)
  	   if count_col eq 0 then ind_header = n_elements(file)-1 else ind_header = ind_column[0]-1
  	   for i=0L,ind_header do begin
  	       tmp1 = strlowcase(strsplit(file[i],' '+string(9b),/extract,count=count))
           if arg_present(deteff) then begin
              if tmp1[0] eq '#detectorefficiencies' then begin
                 for j=1,count-1 do begin
                     ind = stregex(tmp1[j],'^spec([0-9]+)=',length=length,/fold_case,/subexpr)
                     if ind[0] ne -1 then deteff[0,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                     ind = stregex(tmp1[j],'^diff([0-9]+)=',length=length,/fold_case,/subexpr)
                     if ind[0] ne -1 then deteff[1,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                 endfor
              endif
           endif
           if arg_present(maspacing) and (count gt 1) then begin
              if tmp1[0] eq '#monospacing' then maspacing[0] = float(tmp1[1]) 
              if tmp1[0] eq '#anaspacing'  then maspacing[1] = float(tmp1[1]) 
           endif
           if arg_present(npoints) and (count gt 1) then begin
              if tmp1[0] eq '#npoints' then npoints = long(tmp1[1]) 
           endif 
           if arg_present(latticeparm) and (count gt 6) then begin
              if tmp1[0] eq '#lattice' then latticeparm[*] = float(tmp1[1:6]) 
           endif
           if arg_present(latticeori) and (count gt 6) then begin
              if tmp1[0] eq '#orient' then latticeori[*] = float(tmp1[1:6]) 
           endif
           if arg_present(instrname) and (count gt 1) then begin
              if tmp1[0] eq '#instrname' then instrname = tmp1[1]
           endif
           if arg_present(comment) and (count gt 1) then begin
              marker = '#comment'
              if tmp1[0] eq marker then comment = strtrim(strmid(file[i],strpos(strlowcase(file[i]),marker)+strlen(marker)),2)
           endif
           if arg_present(temperatureunit) and (count gt 1) then begin
              if tmp1[0] eq '#temperatureunits' then temperatureunit = tmp1[1]
           endif
           if (tmp1[0] eq '#filename') and (count gt 1) then filename = tmp1[1]
  	   endfor
  	   header = file[0:ind_header]
  	   if count_col gt 0 then begin
  	      tmp_info = strsplit(file[ind_column[0]],' '+string(9b),/extract,count=n_tmp1)
  	      tmp1 = strlowcase(tmp_info)
          n_info = n_elements(info)
          if n_info eq 0 then begin
             info   = tmp1[1:(n_tmp1-1)]
             n_info = n_tmp1-1
          endif else if n_info ne (n_tmp1-1) then begin
             ok = dialog_message('The files are of different format',/center,/error,dialog_parent=group_leader)
             error = 1b & return
          endif
          ind_spec = where(stregex(info,'spec[0-2][0-9]',/boolean))
          ind_diff = where(stregex(info,'diff[0-2][0-9]',/boolean))
          if arg_present(flipstring) then indflip = (where(info eq 'flip',check_flip))[0]
          for i=ind_column[0]+1,n_elements(file)-1 do begin
              if stregex(file[i],'^[ '+string(9b)+']*#',/boolean) then begin
                 if stregex(file[i],'#ng0:end',/boolean,/fold_case) and n_elements(data) ne 0 then npoints = n_elements(data[0,*]) ;force header npoints to be actual data number since the file is complete 
                 if n_elements(trailer) eq 0 then trailer = file[i] else trailer = [trailer,file[i]]
                 continue
              endif
              if stregex(file[i],'^[ '+string(9b)+']*$',/boolean) then continue
              tmp2 = strsplit(file[i],' '+string(9b),/extract,count=n_tmp1)
              if keyword_set(string) then begin
                 if n_info ne n_tmp1 then begin
                    if n_tmp1 gt n_info then begin
                       if dm_to_number(tmp2[n_tmp1-1],/double) gt 1243526755.85 then tmp2 = [tmp2[0:(n_info-2)],tmp2[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp2 = tmp2[0:(n_info-1)]
                    endif else begin
                       if dm_to_number(tmp2[n_tmp1-1],/double) gt 1243526755.85 then tmp2 = [tmp2,replicate('',n_info-n_tmp1-1),tmp2[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp2 = [tmp2,replicate('',n_info-n_tmp1)]
                    endelse
                 endif
                 if n_elements(data) eq 0 then data = tmp2 else data = [[data],[tmp2]]
              endif else begin             
                 tmp1 = dm_to_number(tmp2,zerostring=zerostring,onestring=onestring,twostring=twostring,/double)
                 if (~keyword_set(checkonly)) and (total(tmp1[[ind_spec,ind_diff]] lt 0) eq 40) then begin ;-1 in all diff and spec counts, ignore them
                    all_neg = 1b
                    continue  
                 endif
                 if n_info ne n_tmp1 then begin
                    if keyword_set(checkonly) then continue
                    if n_elements(inc_ans) eq 0 then begin
                       if keyword_set(answeryes) or keyword_set(answerno) then inc_ans = (['yes','no'])[keyword_set(answerno)] $
                       else inc_ans = dialog_message([filename+' has inconsistent number of columns.','Do you want to continue?','','If yes, all missing data will be set to 0.'],/default_no,/question,/center,dialog_parent=group_leader,title='Please confirm:')
                       if strmatch(inc_ans,'no',/fold_case) then begin
                          error = 1b & free_lun,unit & return
                       endif
                    endif
                    if n_tmp1 gt n_info then begin
                       if tmp1[n_tmp1-1] gt 1243526755.85 then tmp1 = [tmp1[0:(n_info-2)],tmp1[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp1 = tmp1[0:(n_info-1)]
                    endif else begin
                       if tmp1[n_tmp1-1] gt 1243526755.85 then tmp1 = [tmp1,replicate(0,n_info-n_tmp1-1),tmp1[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp1 = [tmp1,replicate(0,n_info-n_tmp1)]
                    endelse
                 endif  
                 if n_elements(data) eq 0 then data = tmp1 else data = [[data],[tmp1]]
                 if keyword_set(check_flip) then begin
                    if n_elements(tmp2) gt indflip then begin
                       tmp3 = round(tmp1[indflip])
                       if (tmp3 ne 0) and (tmp3 ne 1) then tmp3 = 2
                       flipstring[tmp3] = tmp2[indflip]
                    endif
                 endif
              endelse
              if keyword_set(firstlineonly) then break
          endfor
  	   endif
  	   if n_elements(comment) ne 0 then begin
          if (strmid(comment,0,1) eq '"') and (strmid(comment,0,1,/reverse_offset) eq '"') then comment = strmid(comment,1,strlen(comment)-2)
       endif
  	   if n_elements(data) ne 0 then data = transpose(data)
       all_neg = keyword_set(all_neg) and (n_elements(data) eq 0)
       if keyword_set(string) then info = tmp_info[1:n_info]
  	   return
  	endif
  	
    for i=0,n_elements(file)-1 do begin
        openr,unit,file[i],/get_lun
        tmp1 = temporary(inc_ans) ;each file will be asked for inconsistent columns
        if eof(unit) then continue
        while(~stregex(tmp,'#columns',/boolean,/fold_case)) do begin
            readf,unit,tmp
            if (i eq 0) and (arg_present(maspacing) or arg_present(latticeparm) or arg_present(latticeori) or arg_present(deteff) or arg_present(instrname) or arg_present(comment) or $
               arg_present(npoints) or arg_present(temperatureunit)) then begin
               tmp1 = strlowcase(strsplit(tmp,' '+string(9b),/extract,count=count))
               if arg_present(deteff) then begin
                  if tmp1[0] eq '#detectorefficiencies' then begin
                     for j=1,count-1 do begin
                         ind = stregex(tmp1[j],'^spec([0-9]+)=',length=length,/fold_case,/subexpr)
                         if ind[0] ne -1 then deteff[0,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                         ind = stregex(tmp1[j],'^diff([0-9]+)=',length=length,/fold_case,/subexpr)
                         if ind[0] ne -1 then deteff[1,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                     endfor
                  endif
               endif
               if arg_present(maspacing) and (count gt 1) then begin
                  if tmp1[0] eq '#monospacing' then maspacing[0] = float(tmp1[1]) 
                  if tmp1[0] eq '#anaspacing'  then maspacing[1] = float(tmp1[1]) 
               endif
               if arg_present(npoints) and (count gt 1) then begin
                  if tmp1[0] eq '#npoints' then npoints = long(tmp1[1])
               endif
               if arg_present(latticeparm) and (count gt 6) then begin
                  if tmp1[0] eq '#lattice' then latticeparm[*] = float(tmp1[1:6]) 
               endif
               if arg_present(latticeori) and (count gt 6) then begin
                  if tmp1[0] eq '#orient' then latticeori[*] = float(tmp1[1:6]) 
               endif
               if arg_present(instrname) and (count gt 1) then begin
                  if tmp1[0] eq '#instrname' then instrname = tmp1[1]
               endif
               if arg_present(comment) and (count gt 1) then begin
                  marker = '#comment'
                  if tmp1[0] eq marker then comment = strtrim(strmid(tmp,strpos(strlowcase(tmp),marker)+strlen(marker)),2)
               endif
               if arg_present(temperatureunit) and (count gt 1) then begin
                  if tmp1[0] eq '#temperatureunits' then temperatureunit = tmp1[1]
               endif
            endif
            if i eq 0 then header = [header,strtrim(tmp,2)]
        endwhile
        if i eq 0 then begin
           header = header[1:n_elements(header)-2]
           filename = file_basename(file[i])
        endif else filename = [filename,file_basename(file[i])]
        if stregex(filename[i],'^ftptmp',/boolean,/fold_case) then filename[i] = strmid(filename[i],6)
        tmp_info = strsplit(tmp,' '+string(9b),/extract,count=n_tmp1)
        tmp1 = strlowcase(tmp_info)
        n_info = n_elements(info)
        if n_info eq 0 then begin
           info   = tmp1[1:(n_tmp1-1)]
           n_info = n_tmp1-1
        endif else if n_info ne (n_tmp1-1) then begin
           ok = dialog_message('The files are of different format',/center,/error,dialog_parent=group_leader)
           error = 1b & free_lun,unit & return
        endif
        ind_spec = where(stregex(info,'spec[0-2][0-9]',/boolean))
        ind_diff = where(stregex(info,'diff[0-2][0-9]',/boolean))
        if arg_present(flipstring) then indflip = (where(info eq 'flip',check_flip))[0]
        while (~eof(unit)) do begin
              readf,unit,tmp
              if stregex(tmp,'^[ '+string(9b)+']*#',/boolean) then begin
                 if (i eq 0) then begin
                    if stregex(tmp,'#ng0:end',/boolean,/fold_case) and n_elements(data) ne 0 then npoints = n_elements(data[0,*]) ;force header npoints to be actual data number since the file is complete
                    if n_elements(trailer) eq 0 then trailer = tmp else trailer = [trailer,tmp]
                 endif
                 continue
              endif
              if stregex(tmp,'^[ '+string(9b)+']*$',/boolean) then continue
              tmp2 = strsplit(tmp,' '+string(9b),/extract,count=n_tmp1)
              if keyword_set(string) then begin
                 if n_info ne n_tmp1 then begin
                    if n_tmp1 gt n_info then begin
                       if dm_to_number(tmp2[n_tmp1-1],/double) gt 1243526755.85 then tmp2 = [tmp2[0:(n_info-2)],tmp2[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp2 = tmp2[0:(n_info-1)]
                    endif else begin
                       if dm_to_number(tmp2[n_tmp1-1],/double) gt 1243526755.85 then tmp2 = [tmp2,replicate('',n_info-n_tmp1-1),tmp2[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp2 = [tmp2,replicate('',n_info-n_tmp1)]
                    endelse
                 endif
                 if n_elements(data) eq 0 then data = tmp2 else data = [[data],[tmp2]]
              endif else begin
                 tmp1 = dm_to_number(tmp2,zerostring=zerostring,onestring=onestring,twostring=twostring,/double)
                 if (~keyword_set(checkonly)) and (total(tmp1[[ind_spec,ind_diff]] lt 0) eq 40) then begin ;-1 in all diff and spec counts, ignore them
                    all_neg = 1b
                    continue  
                 endif
                 if n_info ne n_tmp1 then begin
                    if keyword_set(checkonly) then continue
                    if n_elements(inc_ans) eq 0 then begin
                       if keyword_set(answeryes) or keyword_set(answerno) then inc_ans = (['yes','no'])[keyword_set(answerno)] $
                       else inc_ans = dialog_message([filename[i]+' has inconsistent number of columns.','Do you want to continue?','','If yes, all missing data will be set to 0.'],/default_no,/question,/center,dialog_parent=group_leader,title='Please confirm:')
                       if strmatch(inc_ans,'no',/fold_case) then begin
                          error = 1b & free_lun,unit & return
                       endif
                    endif
                    if n_tmp1 gt n_info then begin
                       if tmp1[n_tmp1-1] gt 1243526755.85 then tmp1 = [tmp1[0:(n_info-2)],tmp1[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp1 = tmp1[0:(n_info-1)]
                    endif else begin
                       if tmp1[n_tmp1-1] gt 1243526755.85 then tmp1 = [tmp1,replicate(0,n_info-n_tmp1-1),tmp1[n_tmp1-1]] $  ;last column is epoch, save it
                       else tmp1 = [tmp1,replicate(0,n_info-n_tmp1)]
                    endelse
                 endif
                 if n_elements(data) eq 0 then data = tmp1 else data = [[data],[tmp1]]
                 if keyword_set(check_flip) then begin
                    if n_elements(tmp2) gt indflip then begin
                       tmp3 = round(tmp1[indflip])
                       if (tmp3 ne 0) and (tmp3 ne 1) then tmp3 = 2
                       flipstring[tmp3] = tmp2[indflip]
                    endif
                 endif
              endelse
              if keyword_set(firstlineonly) then break
        endwhile
        free_lun,unit
	  endfor
	  if n_elements(comment) ne 0 then begin
	     if (strmid(comment,0,1) eq '"') and (strmid(comment,0,1,/reverse_offset) eq '"') then comment = strmid(comment,1,strlen(comment)-2)
	  endif
	  if n_elements(data) ne 0 then data = transpose(data)
	  all_neg = keyword_set(all_neg) and (n_elements(data) eq 0)
	  if keyword_set(string) then info = tmp_info[1:n_info]
end

;this procedure reads macs file for dcs_mslice program
;Parameters:
;   open_path:    file path, string
;   open_file:    file name(s), string or string array
;Keywords:
; input:
;   ftpobj:       ftp object for ftp files
;   parent:       message widget parent,  default is self.tlb
;   title:        message title
; output:
;   error:        error flag 1 or 0
;   qty:          [2,ndet,ndat] (0,*,*)-spec (1,*,*) diff
;   dqty:         [2,ndet,ndat]
;   a3:           [ndat] 
;   baddet_id:    an array of the bad detector id. for load vanadium file, no warning given here
;   cfx:          [ndat] flag for cfx monitors 0-none 1-Be 2-PG 3-MgF2 
;   comment:      comment string
;   deteff:       [2,ndet] detector efficiency [0,*] for spec detector, [1,*] for diff detector, note: the efficiency is more like normalizaiton factor Int*deteff, instead of Int/deteff
;   ei:           [ndat]
;   ef:           [ndat]
;   hfield:       [ndat], available when self.extravaxis_yn[1]=1
;   info:         {latticeparm:latticeparm,latticeori:latticeori,temperature:setpoint,magfield:magfield}  used for checking file compatibility
;   kidney:       [ndat] kidney angle
;   montype:      0-time 1-monitor, check only the first file
;   ptai:         [ndat], integer array
;   specify:      [ndat], available when self.extravaxis_yn[2]=1
;   temperature:  [ndat,3]  sample, control, and setpoint temperature, in Kelvins
;   weight:       [ndat,2]  [ndat,0]-time [ndat,1]-monitor
pro dcs_mslice::dm_load_macs,open_path,open_file,a3=a3,baddet_id=baddet_id,cfx=cfx,comment=comment,deteff=deteff,ei=ei,ef=ef,error=error,ftpobj=ftpobj,hfield=hfield,info=info,$
    kidney=kidney,montype=montype,parent=parent,ptai=ptai,qty=qty,dqty=dqty,specify=specify,temperature=temperature,title=title,weight=weight
    cd,current=current
    if n_elements(parent) eq 0 then parent = self.tlb
    if n_elements(title)  eq 0 then title  = 'Loading data ...'
    if n_elements(qty)    ne 0 then tmp    = temporary(qty)   ;destroy qty
    if n_elements(dqty)   ne 0 then tmp    = temporary(dqty)  ;destroy dqty
    checklattuv = self.checklattuv
    n_files = n_elements(open_file)
    cnt_cfx0 = 0L ;number of files with cfx=0
    
    ;catch and clear possible io errors
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       if obj_valid(mesg) then obj_destroy,mesg
       ok = dialog_message(!error_state.msg,dialog_parent=self.tlb,/error,/center)
       cd,current
       error = 1b
       return
    end
    
    if obj_valid(ftpobj) then pathsep = '/' else pathsep = self.pathsep
    while(strmid(open_path,0,1,/reverse_offset) eq  pathsep) do  open_path = strmid(open_path,0,strlen(open_path)-1)
    if self.extravaxis_yn[2] then begin
       tmp_spiden = strlowcase(strtrim(self.extravaxis_spec[2],2))
       tmp_spname = self.extravaxis_spec[0]+((strlen(self.extravaxis_spec[1]) eq 0)?'':(' ('+self.extravaxis_spec[1]+')'))
    endif
    for i=0L,n_files-1L do begin
        file = open_path+pathsep+open_file[i]
        if obj_valid(ftpobj) then begin
           ok = ftpobj->GetFileContent(file,localfilename=self.dirs[2]+self.pathsep+'ftptmp'+open_file[i])
           file = self.dirs[2]+self.pathsep+'ftptmp'+open_file[i]
        endif
        if i eq 0 then begin
           mesg =  obj_new('dm_progress',title=title,message='Loading '+file,group_leader=parent) 
           dm_load_macs,file,data,colinfo,header=header,deteff=deteff,maspacing=maspacing,latticeparm=latticeparm,latticeori=latticeori,group_leader=parent,comment=comment,instrname=instrname,$
              temperatureunit=temperatureunit,error=error
        endif else begin
           mesg->update,message='loading '+file,title='Loading...  '+strtrim(string(i+1),2)+'/'+strtrim(string(n_files),2)
           dm_load_macs,file,data,colinfo,header=header,deteff=deteff,maspacing=maspacing,latticeparm=latticeparm,latticeori=latticeori,group_leader=parent,instrname=instrname,$
              temperatureunit=temperatureunit,error=error
        endelse 
        if obj_valid(ftpobj) then file_delete,file,/ALLOW_NONEXISTENT,/NOEXPAND_PATH,/QUIET

        if error ne 0 then begin
           if obj_valid(mesg) then obj_destroy,mesg
           cd,current
           return
        endif
        
        if n_elements(data) eq 0 then continue

        if (instrname ne 'ng0') and (instrname ne 'bt9') then begin
           exitinfo = open_file[i]+' is not a MACS data file.'
        endif else if n_elements(info) eq 0 then begin
           info  = {latticeparm:latticeparm,latticeori:latticeori}
           file0 = open_file[i]
        endif else if checklattuv then begin
           if (total(abs(info.latticeparm-latticeparm)) ne 0) or (total(abs(info.latticeori-latticeori)) ne 0) then begin
              exitinfo = 'Selected data files are of different lattice parameters or scattering plane.'
              if n_elements(file0) eq 0 then file0 = 'previous'
              tmp = [file0+':  '+dm_to_string(info.latticeparm,sep=', ')+'  '+dm_to_string(info.latticeori,sep=', '),$
                     open_file[i]+':  '+dm_to_string(latticeparm,sep=', ')+'  '+dm_to_string(latticeori,sep=', '),' ']
              ans = dialog_message([exitinfo,tmp,'Do you want to temporarily ignore the differences?'],/default_no,/question,/center,dialog_parent=parent,title='Please confirm:')
              if strmatch(ans,'Yes',/fold_case) then begin
                 checklattuv = 0b
                 tmp = temporary(exitinfo)
              endif
           endif   
        endif
        if n_elements(exitinfo) ne 0 then begin
           if n_elements(ans) eq 0 then ok = dialog_message(exitinfo,/error,/center,dialog_parent=parent)
           cd,current
           if obj_valid(mesg) then obj_destroy,mesg
           error = 1b
           return
        endif
        ;check index every time, since the file format could be different
        ind_qty = lonarr(2,20)
        ind_err = lonarr(20)  ;for polarization corrected file, only for spec detectors
        ind_ana = lonarr(20)
        for j=1,20 do begin
            tmp = string(j,format='(i02)')
            ind_qty[0,j-1] = (where(colinfo eq 'spec'+tmp))[0]
            ind_qty[1,j-1] = (where(colinfo eq 'diff'+tmp))[0]
            ind_err[j-1]   = (where(colinfo eq 'spec'+tmp+'err'))[0]
            ind_ana[j-1]   = (where(colinfo eq 'analyzertheta'+tmp))[0]
        endfor
        readerr = (total(ind_err ge 0) eq 20)   ;flag for reading intensity error bars
        ind_a3  = (where(colinfo eq 'a3'))[0]
        ind_a2  = (where(colinfo eq 'a2'))[0]
        ind_a5  = (where(colinfo eq 'a5'))[0]
        if finite(data[0,ind_a5],/nan) then begin
           tmp_x = reform(data[0,ind_ana]) & tmp_dy = fltarr(20)+1
           dm_step_bin,0.5,tmp_x,yerr=tmp_dy
           tmp = min(tmp_dy,ind_tmp)
           tmp = min(abs(data[0,ind_ana]-tmp_x[ind_tmp]),ind_tmp)
           ind_a5 = ind_ana[ind_tmp]
        endif
        ind_kn  = (where(colinfo eq 'kidney'))[0]
        ind_wt0 = (where(colinfo eq 'time'))[0]
        ind_wt1 = (where(colinfo eq 'monitor'))[0]
        ind_t   = (where(colinfo eq 'temp',cnt_t))[0]
        ind_tc  = (where(colinfo eq 'temperaturecontrolreading',cnt_tc))[0]
        ind_ts  = (where(colinfo eq 'temperaturesetpoint',cnt_ts))[0]
        ind_pt  = (where(colinfo eq 'ptai',cnt_pt))[0]
        nstr    = 'cfx'+['be','hopg','mgf']
        cnt_cfx = 0
        for j=0,n_elements(nstr)-1 do begin
            ind_cfx = (where(colinfo eq nstr[j],count))[0]
            if count ne 0 then begin
               if finite(data[0,ind_cfx]) then begin
                  if round(data[0,ind_cfx]) then begin
                     cnt_cfx = cnt_cfx+2^j
                  endif
               endif
            endif
        endfor
        ;check if the date is after CFXdate
        tmp = stregex(header,'#Date(.*)',/fold_case,/subexpr,length=len)
        ind = where(len[1,*] gt 0,count)
        if count gt 0 then begin
           date = dm_to_number(strmid(header[ind[0]],tmp[1,ind[0]],len[1,ind[0]]),/date)
           if (date lt self.macs_CFXdate[1]) and (cnt_cfx eq 4) then cnt_cfx = 0
           if (cnt_cfx eq 0) and (date ge self.macs_CFXdate[0]) then cnt_cfx0 = cnt_cfx0+1
        endif
        ;temperature
        if cnt_t ne 0 then begin
           thistemperature = data[*,[ind_t,ind_tc,ind_ts]]
           if self.tempavg_yn then begin
              thistemperature = (fltarr(n_elements(data[*,0]))+1.)#total(thistemperature,1)/n_elements(thistemperature[*,0])  ;average temperature of the file
           endif
           ;convert all temperature units to Kelvin
           if n_elements(temperatureunit) ne 0 then begin
              case temperatureunit of 
                   'celsius' :   thistemperature = thistemperature+273.15
                   'fahrenheit': thistemperature = (thistemperature+459.67)*5./9.
                   else:
              endcase
           endif
        endif else begin
           if self.extravaxis_yn[0] then begin
              repeat begin
                  ans0 = dm_dialog_input('T (K)',/float,default=ans0,dialog_parent=self.tlb,/align_center,info=open_file[i]+" doesn't contain temperature. Please enter it.",title='Missing temperature',cancel=cancel)
                  if keyword_set(cancel) then message,'Missing temperature value. Go to Option->Viewing Axis->Allow Extra Viewing Axis to remove the temperature axis.',level=0,/noprint,/noname
              endrep until finite(ans0)
              thistemperature = ans0+fltarr(n_elements(data[*,0]),3)
           endif else $
              thistemperature = -1.0+fltarr(n_elements(data[*,0]),3)
        endelse      
        
        ;hfield
        ind_hf = (where(colinfo eq 'magfield',cnt_hf))[0]
        if cnt_hf eq 0 then ind_hf = (where(colinfo eq 'hfield',cnt_hf))[0]
        if self.extravaxis_yn[1] then begin
           if cnt_hf eq 0 then begin
              repeat begin
                  ans1 = dm_dialog_input('H (T)',/float,default=ans1,dialog_parent=self.tlb,/align_center,info=open_file[i]+" doesn't contain magnetic field. Please enter it.",title='Missing magnetic field',cancel=cancel)
                  if keyword_set(cancel) then message,'Missing magnetic field value. Go to Option->Viewing Axis->Allow Extra Viewing Axis to remove the magnetic field axis.',level=0,/noprint,/noname
              endrep until finite(ans1)
              this_hfield = ans1+fltarr(n_elements(data[*,0]))
           endif else this_hfield = data[*,ind_hf]
        endif
        
        ;specified quantity
        if self.extravaxis_yn[2] then begin
           if strlen(tmp_spiden) ne 0 then ind_sp = (where(colinfo eq tmp_spiden,cnt_sp))[0] $
           else cnt_sp = 0
           if cnt_sp eq 0 then begin
              repeat begin
                  ans2 = dm_dialog_input(tmp_spname,/float,default=ans2,dialog_parent=self.tlb,/align_center,info=open_file[i]+" doesn't contain "+self.extravaxis_spec[0]+'. Please enter it.',title='Missing '+self.extravaxis_spec[0],cancel=cancel)
                  if keyword_set(cancel) then message,'Missing '+self.extravaxis_spec[0]+' value. Go to Option->Viewing Axis->Allow Extra Viewing Axis to remove the extra axis.',level=0,/noprint,/noname
              endrep until finite(ans2)
              this_spec = ans2+fltarr(n_elements(data[*,0]))
           endif else this_spec = data[*,ind_sp]
        endif
        
        if cnt_ts ne 0 then begin
           tnames = tag_names(info)
           index  = where(tnames eq 'TEMPERATURE',count)
           if count eq 0 then info = create_struct('TEMPERATURE',data[n_elements(data[*,0])-1,ind_ts],info)   ;use the last setpoint temperature   
        endif
        
        if cnt_hf ne 0 then begin
           tnames = tag_names(info)
           index  = where(tnames eq 'MAGFIELD',count)
           if count eq 0 then info = create_struct('MAGFIELD',data[0,ind_hf],info)
        endif
        
        ;some analyzers may be off, needs to check
        for j=0L,n_elements(data[*,0])-1 do begin
            min_da5 = min(abs(data[j,ind_a5]-data[j,ind_ana]))
            bad_id  = where(abs(data[j,ind_a5]-data[j,ind_ana]) gt min_da5+0.5,badcount,complement=good_id)   ;more than 0.5 degree off
            if badcount ne 0 then begin
               data[j,ind_a5] = mean(data[j,ind_ana[good_id]])
               if ~self.ftoolflag[3] then begin
                  if n_elements(baddetid) eq 0 then baddetid = bad_id else begin
                     baddetid = [baddetid,bad_id]
                     baddetid = baddetid[uniq(baddetid,sort(baddetid))] 
                  endelse 
               endif
            endif
        endfor
        
        if n_elements(qty) eq 0 then begin  
           qty = transpose([[[data[*,ind_qty[0,*]]]],[[data[*,ind_qty[1,*]]]]])
           if readerr then dqty = transpose([[[data[*,ind_err]]],[[1.0>(sqrt(data[*,ind_qty[1,*]]))]]])
           a3  = data[*,ind_a3]
           ei  = 81.8042/((2*maspacing[0]*sin(!dtor*data[*,ind_a2]/2))^2)   ;A2->Ei
           ef  = 81.8042/((2*maspacing[1]*sin(!dtor*data[*,ind_a5]))^2)     ;A5->Ef
           ptai = round(data[*,ind_pt])
           kidney = data[*,ind_kn]
           weight = transpose(data[*,[ind_wt0,ind_wt1]])
           temperature = thistemperature         
           cfx = (intarr(n_elements(data[*,0]))+1)*cnt_cfx
           if self.extravaxis_yn[1] then hfield = this_hfield
           if self.extravaxis_yn[2] then specify = this_spec
        endif else begin
           tmp = transpose([[[data[*,ind_qty[0,*]]]],[[data[*,ind_qty[1,*]]]]])
           qty = [[[qty]],[[tmp]]]
           if readerr then begin
              if n_elements(dqty) eq 0 then dqty = (1.0>(sqrt(qty)))
              tmp  = transpose([[[data[*,ind_err]]],[[1.0>(sqrt(data[*,ind_qty[1,*]]))]]])
              dqty = [[[dqty]],[[tmp]]]
           endif else if n_elements(dqty) ne 0 then begin
              dqty = [[[dqty]],[[1.0>(sqrt(tmp))]]]
           endif
           a3  = [a3,data[*,ind_a3]]
           ei  = [ei,81.8042/((2*maspacing[0]*sin(!dtor*data[*,ind_a2]/2))^2)]
           ef  = [ef,81.8042/((2*maspacing[1]*sin(!dtor*data[*,ind_a5]))^2)]
           ptai = [ptai,round(data[*,ind_pt])]
           kidney = [kidney,data[*,ind_kn]]
           weight = [[weight],[transpose(data[*,[ind_wt0,ind_wt1]])]]    
           temperature = transpose([[transpose(temperature)],[transpose(thistemperature)]])         
           cfx = [cfx,(intarr(n_elements(data[*,0]))+1)*cnt_cfx]
           if self.extravaxis_yn[1] then hfield = [hfield,this_hfield]
           if self.extravaxis_yn[2] then specify = [specify,this_spec]
        endelse
        ;check monitor type
        if n_elements(montype) eq 0 then begin
           yn = stregex(header, '#reference',/boolean,/fold_case)
           index = where(yn,count)
           if count ne 0 then begin
              tmp = strlowcase(strsplit(header[index[0]],' '+string(9b),/extract))
              if tmp[n_elements(tmp)-1] eq 'time'    then montype = 0
              if tmp[n_elements(tmp)-1] eq 'monitor' then montype = 1
           endif
           if n_elements(montype) eq 0 then begin ;in case '#reference' info is absent
              tmp = [total(abs(weight[0,*]-weight[0,0])),total(abs(weight[1,*]-weight[1,0]))]
              tmp = min(tmp,montype)
           endif
        endif
        if (~ obj_valid(mesg)) then break
    endfor

    ;check for NAN's in Ei and Ef, if found, replace them with the corresponding values in Ef and Ei
    if n_elements(qty) gt 0 then begin
       ind = where(finite(ei,/nan),count)
       if count gt 0 then ei[ind] = ef[ind]
       ind = where(finite(ef,/nan),count)
       if count gt 0 then ef[ind] = ei[ind]
       if n_elements(dqty) eq 0 then dqty = sqrt(1.0>(qty))
       weight = transpose(weight) ;[2,ndat]->[ndat,2]
    endif

    cd,current
    if obj_valid(mesg) then obj_destroy,mesg

    ;monitor lambda/2 correction warning
    if cnt_cfx0 ne 0 and ~self.macs_lamb2 then begin
       if n_files eq 1 then mesg = 'No CFX filter is used in the file.' $
       else begin
          if cnt_cfx0 eq n_files then mesg = 'No CFX filter is used in all the files.' $
          else mesg = 'No CFX filter is used in some of the files.'
       endelse
       ok = dialog_message(dialog_parent=self.tlb,mesg+' Do you want to enable the monitor lambda/2 correction in the Option menu?',/question,/center,title='Monitor lambda/2 correction')   
       if ok eq 'Yes' then begin
          self.macs_lamb2 = 1b
          dm_set_button,widget_info(self.tlb,find_by_uname='monlamb2'),self.macs_lamb2,onstring='Apply Monitor Lambda/2 Correction'
       endif
    endif
    
    ;bad detector warning
    nbd = n_elements(baddetid)
    if nbd ne 0 then begin
       if arg_present(baddet_id) then begin ;for load vanadium file, warning will be given there
          baddet_id = baddetid
          return 
       endif
       if ptr_valid(self.mask) then begin
          if self.masktype eq 0 then begin
             if n_elements(*(self.mask)) ge nbd then begin
                tmp = dm_common(baddetid,*(self.mask),count)
                if count eq nbd then nodetwarning = 1b          ;mask already exists, no need to warn
             endif
          endif
       endif
       if ~keyword_set(nodetwarning) then begin
          ok = dialog_message(dialog_parent=self.tlb,['The anaylzer angle A5 in detector '+dm_to_string(baddetid,sep=', ')+([' are',' is'])[nbd eq 1]+' more than 0.5 degree off.',$
               'Do you want to mask '+(['them','it'])[nbd eq 1]+'?','','(Detector number starts from 0.)'],/question,/center,title='mask detector')
          if ok eq 'Yes' then begin
             if ptr_valid(self.mask) and self.masktype eq 0 then begin
                baddetid = [baddetid,*(self.mask)]
                baddetid = baddetid[uniq(baddetid,sort(baddetid))] 
             endif else self.masktype = 0
             self->clear_mask,/keep_ar
             self.mask = ptr_new(baddetid)
             self->my_widget_control,['mksvBut','mkclBut'],/sensitive
             cid = widget_info(self.tlb,find_by_uname='mkstdnum') 
             if cid ne 0 then self->event,{WIDGET_BUTTON,ID:cid,TOP:self.tlb,HANDLER:self.tlb,SELECT:1}
          endif else self.ftoolflag[3] = 1b   ;no warning from here on, clear data will clear this flag
       endif
    endif else self.ftoolflag[3] = 0b         ;reset the warning flag
end